/* $Id: mwrite.c,v 1.3 2005/02/09 19:28:37 sassek Exp $ */
#ifdef WIN32
#include <windows.h>
#endif 

#ifdef SSA_OS_UNIX
#include <sys/resource.h>
#include <pthread.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "ibase.h"

// Number of tests defined
#define NUM_TESTS 16

// The number of times each test is to be run
#define NUM_LOOPS   100

// Ugly macro wrappers for lock routines
// IGNORE these for now as they are not used !!!
#if 0
#define fbLock                  CRITICAL_SECTION
#define fbInitLock(plock)       (plock ? InitializeCriticalSection(plock) : 0)
#define fbDestroyLock(plock)	  (plock ? DeleteCriticalSection(plock) : 0)
#define fbGetLock(plock)        (plock ? EnterCriticalSection(plock) : 0)
#define fbReleaseLock(plock)    (plock ? LeaveCriticalSection(plock) : 0)
#else

// Stub out the lock calls to do nothing ...
#define fbLock                  int
#define fbInitLock(plock)       
#define fbDestroyLock(plock)	
#define fbGetLock(plock)
#define fbReleaseLock(plock)

#endif

// These arguments are passed to each Threads
typedef struct _ThreadArgs {
	ISC_STATUS_ARRAY    status;     // status array passed to FB functions
	int                 threadID;   // The number for this thread
   int                 tableID;    // The table number for this thread to access
   fbLock              *plock;     // pointer to the lock for connection serialization
   int                 doCreates;  // Should this thread create it's table?
   int                 doFetch;    // Should this thread fetch?
	int                 done;       // Indicates the thread is finished
   const char          *database;  // database to be used
   int testcase; // value from 1-NUM_TESTS, determines test case to run for this thread
   int iterations; // number of times to execute this testcase
} ThreadArgs, *Threadargs;


// ------------------------------------------------------------
// There should be no need to change these
#define USERID      "sysdba"
#define PASSWORD    "masterkey"
#define CHARSET     "NONE"

// Ugly macro wrapper for create thread routine
#ifdef WIN32
#define fbCreateThread(threadfn, arg)   \
    CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)threadfn, arg, 0, NULL);
#endif

#ifdef SSA_OS_UNIX
void fbCreateThread(int (*ThreadFn)(Threadargs args), Threadargs args)
{
    pthread_t thread;
    pthread_attr_t attr;

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    pthread_create(&thread,   &attr, ThreadFn, args);
    
    pthread_attr_destroy(&attr);

}


long GetTickCount()
{
#if 0
     struct rusage r_usage;
     memset(&r_usage, 0, sizeof(r_usage));     
     getrusage(RUSAGE_SELF, &r_usage);
     return r_usage.ru_utime. tv_usec;
#else
     return time(NULL) * 1000;
#endif
}

void Sleep(int sec)
{
    usleep(sec * 1000);
}

#endif



// -----------------------------
// Wrapper functions for FB calls
void fberror(
    Threadargs     args,
    const char      *msg)
{
    printf("ERROR: rc = %d during \"%s\"\n", 
        (int)*args->status, msg);
    isc_print_status ( args->status) ; 
    
    exit(1);
} // error


int fbconnect(
    Threadargs      args,
    const char      *database,
    isc_db_handle   *pdb)
{
    char  dpb_buffer[256];
    char  *dpb = dpb_buffer;
    int   rc;

    // build dbp to connect
    *dpb++ = isc_dpb_version1;

    *dpb++ = isc_dpb_user_name;
    *dpb++ = (char)strlen(USERID);
    memcpy(dpb, USERID, strlen(USERID));
    dpb += strlen(USERID);

    *dpb++ = isc_dpb_password;
    *dpb++ = (char)strlen(PASSWORD);
    memcpy(dpb, PASSWORD, strlen(PASSWORD));
    dpb += strlen(PASSWORD);

    *dpb++ = isc_dpb_lc_ctype;
    *dpb++ = (char)strlen(CHARSET);
    memcpy(dpb, CHARSET, strlen(CHARSET));
    dpb += strlen(CHARSET);
    
    // fbGetLock(args->plock);

    // Should not need semaphore protection
    rc = isc_attach_database(args->status,
                             (short)strlen(database), 
                             (char*)database,          // nice lack of const ...
                             pdb, 
                             (short)(dpb-dpb_buffer), 
                             dpb_buffer);

    // fbReleaseLock(args->plock);

    return rc;
} // connect


int fbdisconnect(
    Threadargs     args,
    isc_db_handle   *pdb)
{
    int rc;

    // fbGetLock(args->plock);

    // Should not need semaphore protection
    rc = isc_detach_database(args->status, pdb);

    // fbReleaseLock(args->plock);

    return rc;
} // disconnect


int fbbeginTran(
    Threadargs     args,
    isc_db_handle   *pdb,
    isc_tr_handle   *ptrans)
{
    char    tpb[8];
    size_t  tpbL = 0;
    int     rc;

// started with:
//    tpb[tpbL++] = isc_tpb_read_committed;
//    tpb[tpbL++] = isc_tpb_consistency;
   tpb[tpbL++] = isc_tpb_version3;
	tpb[tpbL++] = isc_tpb_concurrency;
   tpb[tpbL++] = isc_tpb_write;
   tpb[tpbL++] = isc_tpb_wait;

// removed to simulate JDBC TRANSACTION_SERIALIZABLE
//    tpb[tpbL++] = isc_tpb_rec_version;

    fbGetLock(args->plock);

    // Lock db connection for multithreaded access
    rc = isc_start_transaction(args->status,
                                 ptrans,
                                 1, pdb, tpbL, tpb);

    fbReleaseLock(args->plock);

    return rc;
} // beginTran


int fballocStmt(
    Threadargs     args,
    isc_db_handle   *pdb,
    isc_stmt_handle *pstmt)
{
    int rc;

    fbGetLock(args->plock);

    // Lock db connection for multithreaded access
    rc = isc_dsql_allocate_statement(
                    args->status, pdb, pstmt);

    fbReleaseLock(args->plock);

    return rc;
} // allocStmt


int fbfreeStmt(
    Threadargs     args,
    isc_stmt_handle *pstmt,
    unsigned short  option)
{
    int rc;

    fbGetLock(args->plock);

    // Lock db connection for multithreaded access
    rc = isc_dsql_free_statement(
                    args->status, pstmt, option);

    fbReleaseLock(args->plock);

    return rc;
} // freeStmt


int fbprepexec(
    Threadargs     args,
    isc_tr_handle   *ptrans,
    isc_stmt_handle *pstmt,
    const char      *sql)
{
    int rc;

    fbGetLock(args->plock);

    // Lock db connection for multithreaded access
    rc = isc_dsql_prepare(args->status, ptrans, pstmt,
                          (unsigned short)strlen(sql), 
                          (char*)sql,          // nice lack of const ...
                          SQL_DIALECT_V6, NULL);

    if (!rc)
        rc = isc_dsql_execute(args->status, ptrans, pstmt, 
                              SQL_DIALECT_V6, // wouldn't this be implied by the prepare ?!?!
                              NULL);

    fbReleaseLock(args->plock);

    return rc;
} // prepexec


int fbprep(
    Threadargs     args,
    isc_tr_handle   *ptrans,
    isc_stmt_handle *pstmt,
    const char      *sql)
{
    int rc;

    fbGetLock(args->plock);

    // Lock db connection for multithreaded access
    rc = isc_dsql_prepare(args->status, ptrans, pstmt,
                          (unsigned short)strlen(sql), 
                          (char*)sql,          // nice lack of const ...
                          SQL_DIALECT_V6, NULL);

    fbReleaseLock(args->plock);

    return rc;
} // prepexec


int fbexeci(
    Threadargs     args,
    isc_db_handle   *pdb,
    isc_tr_handle   *ptrans,
    const char      *sql)
{
    int rc;

    fbGetLock(args->plock);

    // Lock db connection for multithreaded access
    rc = isc_dsql_execute_immediate(args->status, 
                                    pdb,
                                    ptrans,
                                    (unsigned short)strlen(sql), 
                                    (char*)sql,          // nice lack of const ...
                                    SQL_DIALECT_V6,
                                    NULL);

    fbReleaseLock(args->plock);

    return rc;
} // execi


int fbcommit(
    Threadargs     args,
    isc_tr_handle   *ptrans)
{
    int rc;

    fbGetLock(args->plock);

    // Lock db connection for multithreaded access (?!?)
    rc = isc_commit_transaction(args->status, ptrans);

    fbReleaseLock(args->plock);

    return rc;
} // commit


int fbrollback(
    Threadargs     args,
    isc_tr_handle   *ptrans)
{
    int rc;

    fbGetLock(args->plock);

    // Lock db connection for multithreaded access (?!?)
    rc = isc_rollback_transaction(args->status, ptrans);

    fbReleaseLock(args->plock);

    return rc;
} // rollback


// Free memory associated with an sqlda struct
void fbfreeSqlda(
    Threadargs     args,
    XSQLDA          **psqlda)
{
    int     i;
    XSQLVAR *var;

    if (!psqlda || !*psqlda)
        return;

    // Free the sqlvar memory
    for (var = (*psqlda)->sqlvar, i = 0; 
        i < (*psqlda)->sqln; 
        var++, i++)
    {
        if (var->sqldata)
            free(var->sqldata);
        if (var->sqlind)
            free(var->sqlind);
    }

    free(*psqlda);
    *psqlda = NULL;

} // freeSqlda


// Allocate an sqlda struct for the associated prepared statement
int fballocSqlda(
    Threadargs     args,
    isc_stmt_handle *pstmt,
    XSQLDA          **psqlda)
{
    if (!psqlda)
        fberror(args, "invalid psqlda specified");

    // Clean up existing SQLDA
    fbfreeSqlda(args, psqlda);

    // Allocate a new sqlda
    *psqlda = malloc(XSQLDA_LENGTH(1));
    if (!*psqlda)
        fberror(args, "malloc failed");

    memset(*psqlda, 0x00, XSQLDA_LENGTH(1));
    (*psqlda)->version = SQLDA_VERSION1;
    (*psqlda)->sqln = 1;

    if (isc_dsql_describe(args->status, pstmt, SQLDA_VERSION1, *psqlda))
        fberror(args, "isc_dsql_describe failed");

    // Alloc this sqlda larger if there are more cols than allocated ...
    if ((*psqlda)->sqld > (*psqlda)->sqln)
    {
        int cnt = (*psqlda)->sqld;

        *psqlda = realloc(*psqlda, XSQLDA_LENGTH(cnt));
        if (!*psqlda)
            fberror(args, "malloc failed");

        memset(*psqlda, 0x00, XSQLDA_LENGTH(cnt));
        (*psqlda)->version = SQLDA_VERSION1;
        (*psqlda)->sqln = cnt;

        // Now, fill it in completely
        if (isc_dsql_describe(args->status, pstmt, SQLDA_VERSION1, *psqlda))
            fberror(args, "isc_dsql_describe failed");

        // Size must be the same as previously described
        if ((*psqlda)->sqld != cnt)
            fberror(args, "isc_dsql_describe returned 2 different column counts");
    }

    // Allocate buffers for data and indicators ...
    if ((*psqlda)->sqld)
    {
        int     i;
        XSQLVAR *var;

        // Alloc the sqlvar memory
        for (var = (*psqlda)->sqlvar, i = 0; 
            i < (*psqlda)->sqln; 
            var++, i++)
        {
            // Reserve space for VARYING len bytes
            var->sqldata = (char*)malloc(var->sqllen + 2);
            var->sqlind  = (short*)malloc(sizeof(*(var->sqlind)));
            if (!var->sqldata || !var->sqlind)
                fberror(args, "malloc failed");
        }

    } // if result set

    // we made it here so it must be OK ...
    return 0;

} // allocSqlda


int fbfetch(
    Threadargs     args,
    isc_stmt_handle *pstmt,
    XSQLDA          *sqlda)
{
    int rc;

    fbGetLock(args->plock);

    // Lock db connection for multithreaded access (?!?)
    rc = isc_dsql_fetch(args->status, pstmt,
                          SQLDA_VERSION1, sqlda);

    fbReleaseLock(args->plock);

    return rc;
} // fetch


char * vstrcpy (char * s, PARAMVARY * vs) {
	// char s[];
	// s = (char *) malloc(vs->vary_length + 1);
	strncpy (s, vs->vary_string, vs->vary_length);
	s[vs->vary_length] = '\0';
	return s;
}


//
// testTransactionsSerializableSelectUpdateWithArithmetic
int mwrite1 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;

   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");

   for (row = 0; row < args->iterations; row++) {
   	int had_error = 0; 
      if (fbbeginTran(args, &db1, &tr1))
	      fberror(args, "beginTran");
                
		sprintf (sql,
			"UPDATE TT SET DOLLARS = DOLLARS + 5 WHERE ANUM = 1;" );
      rc = fbexeci(args, &db1, &tr1, sql);
      if (rc) {
      	had_error = 1; 
      	isc_print_status(args->status);
			if (fbrollback(args, &tr1)) {
				isc_print_status(args->status);
				printf("MWRITE1 ERROR ON ROLLBACK - shouldn't happen.\n");
			}
      }
      
      if (had_error == 0) {
			sprintf (sql,
				"UPDATE TT SET DOLLARS = DOLLARS -5 WHERE ANUM = 1;" );
	      rc = fbexeci(args, &db1, &tr1, sql);
	      if (rc) {
	      	had_error=1; 
	      	isc_print_status(args->status);
				if (fbrollback(args, &tr1)) {
					isc_print_status(args->status);
					printf("MWRITE1 ERROR ON ROLLBACK - shouldn't happen.\n");
				}
	      } 
      }
      
      if (had_error == 0) {
		   if (fbcommit(args, &tr1)) {
				isc_print_status(args->status);
				// this error message is OK, even expected   	
				// GDS Exception. 335544336. deadlock\n
				// update conflicts with concurrent update
				if (fbrollback(args, &tr1)) {
					isc_print_status(args->status);
					printf("MWRITE1 ERROR ON ROLLBACK - shouldn't happen.\n");
				}
			}
		}
      	
	} // for row 

   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 
	return 0; 

} // 


//
//
int mwrite2 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;

   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");

   for (row = 0; row < args->iterations; row++) {
		int had_error=0;
		
      if (fbbeginTran(args, &db1, &tr1))
	      fberror(args, "beginTran");
                
		sprintf (sql,
			"UPDATE TT SET DOLLARS = DOLLARS + 10 WHERE ANUM = 1;" );
      rc = fbexeci(args, &db1, &tr1, sql);
      if (rc) {
      	had_error = 1; 
      	isc_print_status(args->status);
			if (fbrollback(args, &tr1)) {
				isc_print_status(args->status);
				printf("MWRITE2 ERROR ON ROLLBACK - shouldn't happen.\n");
			}
      }
      
      if (had_error == 0) {
			sprintf (sql,
			"UPDATE TT SET DOLLARS = DOLLARS - 10 WHERE ANUM = 1;" );
	      rc = fbexeci(args, &db1, &tr1, sql);
	      if (rc) {
	      	had_error=1; 
	      	isc_print_status(args->status);
				if (fbrollback(args, &tr1)) {
					isc_print_status(args->status);
					printf("MWRITE2 ERROR ON ROLLBACK - shouldn't happen.\n");
				}
	      } 
      }
      
      if (had_error == 0) {
		   if (fbcommit(args, &tr1)) {
				isc_print_status(args->status);
				// this error message is OK, even expected   	
				// GDS Exception. 335544336. deadlock\n
				// update conflicts with concurrent update
				if (fbrollback(args, &tr1)) {
					isc_print_status(args->status);
					printf("MWRITE2 ERROR ON ROLLBACK - shouldn't happen.\n");
				}
			}
		}

	} // for row 

   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

	return 0; 

} // 


// testTransactionSerializabilityDeadlockManagement()
int mwrite3 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row,loop;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;
   int anum; 
   int anum_initial; 
   int had_error = 0; 

   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");

   for (row = 0; row < args->iterations; row++) {

		// try to update aa 50 times
      for (loop = 0; loop < 50; loop++) {
      	had_error=0; 
      	if (fbbeginTran(args, &db1, &tr1))
   	   	fberror(args, "beginTran1");

   		sprintf (sql, "UPDATE aa SET anum = anum + 3;" );
   		if (fbprepexec(args, &tr1, &stmt1, sql)) {
   			isc_print_status(args->status);
   			loop--;
   			had_error=1; 
   			if (fbrollback(args, &tr1)) {
   				printf ("ERROR ON ROLLBACK!!! This is bad. \n"); 
   				isc_print_status(args->status);
   			}
   		}
   	   if (had_error==0) {
   	   	if (fbcommit(args, &tr1)) {
	   			// this error message is OK, even expected   	
   				// GDS Exception. 335544336. deadlock\n
   				// update conflicts with concurrent update
      			isc_print_status(args->status); 
	   			if (fbrollback(args, &tr1)) {
   					printf ("ERROR ON ROLLBACK!!! This is bad. \n"); 
   					isc_print_status(args->status);
   				}
   			}
   			} else
   				had_error=0;
   		} // for AA
   
		// try to update BB 4 times
      for (loop = 0; loop < 4; loop++) {
      	had_error = 0; 
      	if (fbbeginTran(args, &db1, &tr1))
   	   	fberror(args, "beginTran2");
   		sprintf (sql, "UPDATE BB SET BNUM = BNUM + 5;" );
   		if (fbprepexec(args, &tr1, &stmt1, sql)) {
   			isc_print_status(args->status);
   			loop--;
   			had_error=1; 
   			if (fbrollback(args, &tr1)) {
   				printf ("ERROR ON ROLLBACK!!! This is bad. \n"); 
   				isc_print_status(args->status);
   			}
   		}
   	   if (had_error==0) {
   	   	if (fbcommit(args, &tr1)) {
	   			// this error message is OK, even expected   	
   				// GDS Exception. 335544336. deadlock\n
   				// update conflicts with concurrent update
   	   		printf ("MWRITE3: error on final commit\n"); 
      			isc_print_status(args->status);
	   			if (fbrollback(args, &tr1)) {
   					printf ("ERROR ON ROLLBACK!!! This is bad. \n"); 
   					isc_print_status(args->status);
   				}
   			} else
   				had_error=0;
   		}
   	} // for BB
	} // for
		
   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

	return 0; 

} // 


int mwrite4 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row,loop;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;
   int anum; 
   int anum_initial; 
   int had_error = 0; 

   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");

   for (row = 0; row < args->iterations; row++) {

      for (loop = 0; loop < 50; loop++) {
      	had_error=0;
      	if (fbbeginTran(args, &db1, &tr1))
   	   	fberror(args, "beginTran3");
   		sprintf (sql, "UPDATE bb SET bnum = bnum - 3;" );
   		if (fbprepexec(args, &tr1, &stmt1, sql)) {
   			isc_print_status(args->status);
   			loop--;
   			had_error=1; 
   			if (fbrollback(args, &tr1)) {
   				printf ("ERROR ON ROLLBACK!!! This is bad. \n"); 
   				isc_print_status(args->status);
   			}
   		}
   	   if (had_error==0) {
   	   	if (fbcommit(args, &tr1)) {
      	   	fberror(args, "commit on mwrite3");
      			isc_print_status(args->status); 
   			// this error message is OK, even expected   	
   			// GDS Exception. 335544336. deadlock\n
   			// update conflicts with concurrent update
   			}
   		} else
   				had_error=0;
   	} // for update BB
   
      for (loop = 0; loop < 4; loop++) {
      	had_error=0;
      	if (fbbeginTran(args, &db1, &tr1))
   	   	fberror(args, "beginTran4");
   		sprintf (sql, "UPDATE aa SET aNUM = aNUM - 5;" );
   		if (fbprepexec(args, &tr1, &stmt1, sql)) {
   			had_error=1; 
   			loop--;
   			isc_print_status(args->status);
   			if (fbrollback(args, &tr1)) {
   				printf ("ERROR ON ROLLBACK!!! This is bad. \n"); 
   				isc_print_status(args->status);
   			}
   		}
   	   if (had_error==0) {
   	   	if (fbcommit(args, &tr1)) {
   			// this error message is OK, even expected   	
   			// GDS Exception. 335544336. deadlock\n
   			// update conflicts with concurrent update
      			isc_print_status(args->status); 
	   			if (fbrollback(args, &tr1)) {
   					printf ("ERROR ON ROLLBACK!!! This is bad. \n"); 
   					isc_print_status(args->status);
   				}
   			}
   		} else
   			had_error=0;
   	} // for update AA
	} // for row
	
   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

	return 0; 

} // 

int mwrite5 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row,loop;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;
   int anum; 
   int anum_initial; 
   int had_error = 0; 

   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
again:
   for (row = 0; row < args->iterations; row++) {
      // try the delete until no exception is thrown
      for (loop = 0; loop < 1; loop++) {
      	if (fbbeginTran(args, &db1, &tr1))
   	   	fberror(args, "beginTran1");
   		sprintf (sql, "DELETE FROM AA WHERE ANUM BETWEEN 1 AND 5;" );
   		if (fbprepexec(args, &tr1, &stmt1, sql)) {
   			had_error=1; 
   			loop--;
   			isc_print_status(args->status);
   			if (fbrollback(args, &tr1)) {
   				printf ("ERROR ON ROLLBACK!!! This is bad. \n"); 
   				isc_print_status(args->status);
   			}
   		}
   	   if (had_error==0) {
   	   	if (fbcommit(args, &tr1)) {
      	   	fberror(args, "commit on mwrite3");
      			isc_print_status(args->status); 
   				}
   			} else
   				had_error=0;
   		}
   
   	// try the insert until no exception is thrown
      for (loop = 0; loop < 1; loop++) {
      	if (fbbeginTran(args, &db1, &tr1))
   	   	fberror(args, "beginTran2");
   		sprintf (sql, "INSERT INTO AA VALUES (1);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		sprintf (sql, "INSERT INTO AA VALUES (2);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		sprintf (sql, "INSERT INTO AA VALUES (3);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		sprintf (sql, "INSERT INTO AA VALUES (4);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		sprintf (sql, "INSERT INTO AA VALUES (5);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		if (fbcommit(args,&tr1)) {
     	   	fberror(args, "commit on mwrite5");
     			isc_print_status(args->status); 
     			loop--;
   			}
   		}
   		
      // try the 2nd delete until no exception is thrown
      for (loop = 0; loop < 1; loop++) {
      	if (fbbeginTran(args, &db1, &tr1))
   	   	fberror(args, "beginTran1");
   		sprintf (sql, "DELETE FROM AA WHERE ANUM BETWEEN 6 AND 10;" );
   		if (fbprepexec(args, &tr1, &stmt1, sql)) {
   			isc_print_status(args->status);
   			loop--;
   			had_error=1; 
   			if (fbrollback(args, &tr1)) {
   				printf ("ERROR ON ROLLBACK!!! This is bad. \n"); 
   				isc_print_status(args->status);
   			}
   		}
   	   if (had_error==0) {
   	   	if (fbcommit(args, &tr1)) {
      	   	fberror(args, "commit on mwrite3");
      			isc_print_status(args->status); 
   				}
   			} else
   				had_error=0;
   		}
   
   	// try the insert until no exception is thrown
      for (loop = 0; loop < 1; loop++) {
      	if (fbbeginTran(args, &db1, &tr1))
   	   	fberror(args, "beginTran2");
   		sprintf (sql, "INSERT INTO AA VALUES (6);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		sprintf (sql, "INSERT INTO AA VALUES (7);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		sprintf (sql, "INSERT INTO AA VALUES (8);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		sprintf (sql, "INSERT INTO AA VALUES (9);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		sprintf (sql, "INSERT INTO AA VALUES (10);" );
   		if (fbprepexec(args, &tr1, &stmt1, sql))
   			fberror(args, "insert value 1\n");
   		if (fbcommit(args,&tr1)) {
     	   	fberror(args, "commit on mwrite5");
     			isc_print_status(args->status); 
     			loop--;
   			}
   		}
   		
   	} // outer for
//printf("mwrite5 done... looping\n");	
goto again;
		
   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

			
	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

	return 0; 

} // 


int mwrite6 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;
   int anum; 
   int anum_initial; 
   int had_error = 0; 
   int rowCount=0;
	int selectCount=0;
   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
again:
   for (row = 0; row < args->iterations; row++) {
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
      	
   	// get value of anum
   	sprintf (sql, "SELECT * FROM AA;");  
		if (fbprepexec(args, &tr1, &stmt1, sql)) 
   		fberror (args->status, "mwrite6-fbprepexec"); 
 		if (fballocSqlda(args, &stmt1, &sqlda ))
			fberror(args, "alloc sqlda");
    	rowCount=0;
    	while ((rc = fbfetch(args, &stmt1, sqlda)) == 0) {
     		int anum = *(int *)sqlda->sqlvar[0].sqldata;
     		rowCount++;
   		} // while rows to fetch
   	if (rc != 100)
   		fberror(args, "fetch");                
   
   
   	if (rowCount % 5 != 0)	{
      	// get value of anum
         if (fbfreeStmt(args, &stmt1, DSQL_close))
             fberror(args, "drop statetment");
      	sprintf (sql, "SELECT * FROM AA;");  
   		if (fbprepexec(args, &tr1, &stmt1, sql)) 
      		fberror (args->status, "mwrite6-fbprepexec"); 
       	if (fballocSqlda(args, &stmt1, &sqlda ))
      		fberror(args, "alloc sqlda");
       	rowCount=0;
       	while ((rc = fbfetch(args, &stmt1, sqlda)) == 0) {
        		int anum = *(int *)sqlda->sqlvar[0].sqldata;
      		printf("anum = %d\n", anum);
        		rowCount++;
      		} // while rows to fetch
      	if (rc != 100)
      		fberror(args, "fetch");                
   	}
   
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   
   	selectCount=0;
   	sprintf (sql, "SELECT count(*) FROM aa;");  
		if (fbprepexec(args, &tr1, &stmt1, sql)) 
   		fberror (args->status, "mwrite6-fbprepexec"); 
    	if (fballocSqlda(args, &stmt1, &sqlda ))
   			fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(args, &stmt1, sqlda)) == 0) {
     		selectCount = *(int *)sqlda->sqlvar[0].sqldata;
   		} // while rows to fetch
   	if (rc != 100)
   		fberror(args, "fetch");                
   
   	if (rowCount!=selectCount)
   		printf("walking the rs returns %d rows, select count(*) returns %d. BAD!!!\n", rowCount,selectCount); 
//   	else
//   		printf("walking the rs returns %d rows, select count(*) returns %d. OK\n", rowCount,selectCount); 

   
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   
      if (fbcommit(args, &tr1))
         fberror(args, "commit");
   	  
	} // outer for loop   	  
goto again;

   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

	return 0; 
	} // mwrite6


// testTransactionRollbackSerialization
int mwrite7 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *osqlda = NULL;
   XSQLVAR				*ovar;
   int anum,bnum,anum_initial,bnum_initial=0;
	int countTT;

   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
   for (row = 0; row < args->iterations; row++) {
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   
   	sprintf(sql, "insert into tt values(1,1);"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	printf ("could not insert into TT - rolling back and exiting this thread\n");
      	isc_print_status(args->status);
   		if (fbrollback(args, &tr1)) {
   			printf ("ERROR ON ROLLBACK MWRITE7!!! This is bad. \n"); 
   			isc_print_status(args->status);
   		}
      	goto cleanup;
      }
      	
   	// get value of anum
   	sprintf(sql, "select anum from aa");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(args->status, &stmt1, osqlda)) == 0) {
   		anum_initial = *(int *) osqlda->sqlvar[0].sqldata;
   //  		printf ("anum_initial: %d\n", anum_initial); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc);
   		isc_print_status(args->status);
   		goto rollback;
   	}
   
   	// update anum
   	// steve is getting failures here now
   	sprintf(sql, "UPDATE aa SET aNUM = aNUM + 150;"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
   		isc_print_status(args->status);
   		goto rollback;
      }
    
   	// get value of bnum
   	sprintf(sql,"select bnum from bb");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(args->status, &stmt1, osqlda)) == 0) {
   		bnum_initial = *(int *) osqlda->sqlvar[0].sqldata;
   //  		printf ("bnum_initial: %d\n", bnum_initial); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(args->status, "error on fetch");                
   	}
   
   	// update bnum
   	sprintf(sql, "UPDATE BB SET BNUM = BNUM + 20;"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
   		isc_print_status(args->status);
   		goto rollback;
      }
   	// get value of anum
   	sprintf(sql,"select anum from aa");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(args->status, &stmt1, osqlda)) == 0) {
   		anum = *(int *) osqlda->sqlvar[0].sqldata;
   //  		printf ("anum: %d\n", anum); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(args->status, "error on fetch");                
   	}
   	if (anum != (anum_initial+150)) {
   		printf ("mwrite7 - ERROR : ANUM = %d, not %d\n", anum, (anum_initial+150));
   		goto rollback;
   		}
   	
   	// get value of bnum
   	sprintf(sql,"select bnum from bb");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(args->status, &stmt1, osqlda)) == 0) {
   		bnum = *(int *) osqlda->sqlvar[0].sqldata;
   //  		printf ("bnum: %d\n", bnum); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(args->status, "error on fetch");                
   	}
    
   	if (bnum != (bnum_initial+20)) {
   		printf ("mwrite7 - ERROR : BNUM = %d, not %d\n", bnum, (bnum_initial+20));
   	}
   
   rollback:	
   	if (fbrollback(args, &tr1)) {
   		printf ("ERROR ON ROLLBACK MWRITE7!!! This is bad. \n"); 
   		isc_print_status(args->status);
   	}
   
   	// have to start new transaction after last rollback
   	// get value count(*) from tt where anum=2. should be 0 since we did rollback
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   	sprintf(sql,"SELECT COUNT(*) FROM TT WHERE ANUM = 1;");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(args->status, &stmt1, osqlda)) == 0) {
   		countTT = *(int *) osqlda->sqlvar[0].sqldata;
     		printf ("mwrite7: countTT: %d\n", countTT); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(args->status, "error on fetch");                
   	}
    
   	if (countTT != 0) {
   		printf ("ERROR : countTTT <> 0");
   	}

      if (fbcommit(args, &tr1))
         fberror(args, "commit");
	}

cleanup:
   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 
return 0; 
} // mwrite7


int mwrite8 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *osqlda = NULL;
   XSQLVAR				*ovar;
   int anum_initial,bnum_initial,anum,bnum=0;
	int countTTT =0;
	ISC_STATUS_ARRAY status; 
	
   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
   for (row = 0; row < args->iterations; row++) {
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   
   	sprintf(sql, "insert into ttt values(2,'b');"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	printf ("error on insert into ttt - exiting this thread\n");
      	isc_print_status (args->status); 
   		if (fbrollback(args, &tr1)) {
   			printf ("ERROR ON ROLLBACK MWRITE8!!! This is bad. \n"); 
   			isc_print_status(args->status);
      	}
      	goto cleanup;
      }
      	
   	// get value of bnum
   	sprintf(sql,"select bnum from bb");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		bnum_initial = *(int *) osqlda->sqlvar[0].sqldata;
   //  		printf ("bnum_initial: %d\n", bnum_initial); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(status, "error on fetch");                
   	}
    
   	// update bnum
   	sprintf(sql, "UPDATE BB SET BNUM = BNUM - 150;"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status (args->status);
      	goto rollback;
      	}
   
   	// get value of anum
   	sprintf(sql,"select anum from aa");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		anum_initial = *(int *) osqlda->sqlvar[0].sqldata;
   //  		printf ("anum_initial: %d\n", anum_initial); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(status, "error on fetch");                
   	}
   
   
   	// update anum
   	sprintf(sql, "UPDATE aa SET aNUM = aNUM - 20;"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status (args->status);
      	goto rollback;
      	}
   
   	// get value of bnum
   	sprintf(sql,"select bnum from bb");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		bnum = *(int *) osqlda->sqlvar[0].sqldata;
   //  		printf ("bnum: %d\n", bnum); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(status, "error on fetch");                
   	}
    
   	if (bnum != (bnum_initial-150)) {
   		printf ("ERROR : BNUM = %d, not %d. Exiting thread...\n", bnum, bnum_initial);
   		goto rollback;
   	}
   	
    
   	// get value of anum
   	sprintf(sql,"select anum from aa");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		anum = *(int *) osqlda->sqlvar[0].sqldata;
   //  		printf ("anum: %d\n", anum); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(status, "error on fetch");                
   	}
   	if (anum != (anum_initial-20)) {
   		printf ("ERROR : ANUM = %d, not %d\n", anum, anum_initial);
   	}
   	
   rollback:
   	if (fbrollback(args, &tr1)) {
   		printf ("ERROR ON ROLLBACK MWRITE8!!! This is bad. \n"); 
   		isc_print_status(args->status);
   	}
   	
   
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   	// get value count(*) from ttt where anum=2. should be 0 since we did rollback
   	sprintf(sql,"SELECT COUNT(*) FROM TTT WHERE ANUM = 2;");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		countTTT = *(int *) osqlda->sqlvar[0].sqldata;
     		printf ("mwrite8: countTTT: %d\n", countTTT); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(status, "error on fetch");                
   	}
    
   	if (countTTT != 0) {
   		printf ("ERROR : countTTT <> 0");
   	}
   	
      if (fbcommit(args, &tr1))
         fberror(args, "commit");

		} // for outer loop

cleanup:
   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 
return 0; 
} // mwrite8


// this is a test for generators
int mwrite9 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;
   XSQLDA            *isqlda = NULL;
   int id; 
   ISC_TIMESTAMP isc_ts; 
   XSQLVAR				*ivar;
   time_t now;
   struct tm *now_tm;
   int i;


   int had_error = 0; 
   int rowCount=0;
	int selectCount=0;
	
   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
   for (row = 0; row < args->iterations; row++) {
   
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   
   	for (i=0;i<NUM_LOOPS;i++) {
      	
   	// get value of the generator
   	sprintf (sql, "select gen_id(test_blob_id_gen,1) from onerow;");
      rc = isc_dsql_prepare(args->status, &tr1, &stmt1,
                           (unsigned short)strlen(sql), 
                           (char*)sql,          // nice lack of const ...
                           SQL_DIALECT_V6, NULL);
      if (rc) 
       fberror (args->status,"could not prepare\n") ;
   
    	if (fballocSqlda(args, &stmt1, &sqlda ))
   		fberror(args, "alloc sqlda");
   
      rc = isc_dsql_execute(args->status, &tr1, &stmt1, 
                               SQL_DIALECT_V6, // wouldn't this be implied by the prepare ?!?!
                               sqlda);
   
    	rowCount=0;
    	while ((rc = fbfetch(args, &stmt1, sqlda)) == 0) {
     		id = *(int *)sqlda->sqlvar[0].sqldata;
     		rowCount++;
   		} // while rows to fetch
   	if (rc != 100)
   		fberror(args, "fetch");                
   	
      if (fbfreeStmt(args, &stmt1, DSQL_drop))
          fberror(args, "drop statetment");
   	
   	if (fballocStmt(args, &db1, &stmt1))
      	fberror(args, "alloc statement");
   
   	// now that we have the id, populate the table
   	sprintf (sql, "insert into test_blob (id, ts_field) values (%d, ?)", id);  
      rc = isc_dsql_prepare(args->status, &tr1, &stmt1,
                           (unsigned short)strlen(sql), 
                           (char*)sql,          // nice lack of const ...
                           SQL_DIALECT_V6, NULL);
      if (rc) 
       fberror (args->status,"could not prepare\n") ;
   
   	// allocate the sqlda structure and fill in the timestamp
    	if (fballocSqlda(args, &stmt1, &isqlda ))
   		fberror(args, "alloc isqlda");
   		
   	time(&now); // now is time_t 
   	//printf ("time is %s\n", ctime(&now)); 
   	now_tm = localtime(&now); // now_tm is struct tm
   	// printf( "%.19s \n", asctime( now_tm ) );
   	isc_encode_timestamp(now_tm, &isc_ts);
   	isc_decode_timestamp(&isc_ts, now_tm);
   	// printf( "%.19s \n", asctime( now_tm ) );
   
   	isqlda = (XSQLDA *)malloc(XSQLDA_LENGTH(1));
   	isqlda->version = SQLDA_VERSION1;
   	isqlda->sqln = 1;
   	isqlda->sqld = 1;
   	ivar = &(isqlda->sqlvar[0]);
   	ivar->sqltype = SQL_TIMESTAMP;
   	ivar->sqllen = sizeof(ISC_TIMESTAMP);
   	ivar->sqldata = (char *) malloc(sizeof(ISC_TIMESTAMP));
   	* (ISC_TIMESTAMP*) (ivar->sqldata) = isc_ts;
   	
   //	* (ISC_TIMESTAMP*) (ivar->sqldata) = isc_ts;
   	//ivar = &(sqlda->sqlvar[0]);
   	//ivar->sqldata = (char *) malloc(sizeof(ISC_TIMESTAMP));
   	//* (ISC_TIMESTAMP*) (ivar->sqldata) = isc_ts;
   
      rc = isc_dsql_execute(args->status, &tr1, &stmt1, 
                               SQL_DIALECT_V6,
                               isqlda);
                               
      if (rc)
      	fberror(args, "isc_dql_execute"); 
      	
	  }
   if (fbcommit(args, &tr1))
      fberror(args, "commit");

	} // outer for loop
   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

return 0; 
} // mwrite9





// testTransactionSerializabilityReadDeleteInsert
int mwrite11 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *osqlda = NULL;
   XSQLVAR				*ovar;
   int anum_initial,bnum_initial,anum,bnum=0;
	int countTTT =0;
	ISC_STATUS_ARRAY status; 
	
   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
   for (row = 0; row < args->iterations; row++) {
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   
   	// get value of anum
   	sprintf(sql,"select anum from aa");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		anum_initial = *(int *) osqlda->sqlvar[0].sqldata;
     		// printf ("anum_initial: %d\n", anum_initial); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(status, "error on fetch");                
   	}
   
   	sprintf(sql, "INSERT INTO BB VALUES (1);"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status(args->status); 
      	goto cleanup;
      }
   
   	sprintf(sql, "DELETE FROM AA WHERE ANUM = 1;"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status(args->status); 
      	goto cleanup;
      }
   
   	// get value of bnum
   	sprintf(sql,"select bnum from bb");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		bnum_initial = *(int *) osqlda->sqlvar[0].sqldata;
     		// printf ("bnum_initial: %d\n", bnum_initial); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(status, "error on fetch");                
   	}
   	
   	// try commit, but if it fails, do rollback
   	if (fbcommit(args, &tr1)) {
   		// commit failed...
   		isc_print_status(args->status);
   		if (fbrollback(args, &tr1)) {
   			isc_print_status(args->status);
   			printf ("\tMWRITE8 ROLLBACK #1 FAILED. THIS IS BAD.\n"); 		
   			goto cleanup;
   		}	
   	}
    
   	// For each value of ?? above, move (insert/delete) the value:
   	// update anum
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   	sprintf(sql, "INSERT INTO AA VALUES (%d);", bnum_initial); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status(args->status); 
      	goto cleanup;
      }
   
   	// update bnum
   	sprintf(sql, "DELETE FROM BB WHERE BNUM = %d;", bnum_initial); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status(args->status); 
      	goto cleanup;
      }
   
   	// try commit, but if it fails, do rollback
   	if (fbcommit(args, &tr1)) {
   		// commit failed...
   		isc_print_status(args->status);
   		if (fbrollback(args, &tr1)) {
   			isc_print_status(args->status);
   			printf ("\tMWRITE8 ROLLBACK #2 FAILED. THIS IS BAD.\n"); 		
   			goto cleanup;
   		}
   	}
   	
   
   cleanup:
   
   	if (tr1 !=0) {
   		if (fbrollback(args, &tr1)) {
   			isc_print_status(args->status);
   			printf ("\tMWRITE8 ROLLBACK #3 FAILED. THIS IS BAD.\n"); 		
   			goto cleanup;
   		}
   	}
   
} // outer for loop 

   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

return 0; 
} // mwrite11




int mwrite12 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *osqlda = NULL;
   XSQLVAR				*ovar;
   int anum_initial,bnum_initial,anum,bnum=0;
	int countTTT =0;
	ISC_STATUS_ARRAY status; 
	
   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
   for (row = 0; row < args->iterations; row++) {
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   
   	// get value of bnum
   	sprintf(sql,"select bnum from bb");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		bnum_initial = *(int *) osqlda->sqlvar[0].sqldata;
     		// printf ("bnum_initial: %d\n", bnum_initial); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(status, "error on fetch");                
   	}
    
   	sprintf(sql, "INSERT INTO AA VALUES (6);"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status(args->status); 
      	goto cleanup;
      }
   	sprintf(sql, "DELETE FROM BB WHERE BNUM = 6;"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status(args->status); 
      	goto cleanup;
      }
   
   	// get value of anum
   	sprintf(sql,"select anum from aa");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		anum_initial = *(int *) osqlda->sqlvar[0].sqldata;
     		// printf ("anum_initial: %d\n", anum_initial); 
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("rc: %d\n", rc); 
   		fberror(status, "error on fetch");                
   	}
   
   	// try commit, but if it fails, do rollback
   	if (fbcommit(args, &tr1)) {
   		// commit failed...
   		isc_print_status(args->status);
   		if (fbrollback(args, &tr1)) {
   			isc_print_status(args->status);
   			printf ("\tMWRITE8 ROLLBACK #2 FAILED. THIS IS BAD.\n"); 		
   			goto cleanup;
   		}
   	}
   	
   	// For value of ?? above, move (insert/delete) the value:
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   	// update bnum
   	sprintf(sql, "insert into bb values (%d);", anum_initial); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status(args->status); 
      	goto cleanup;
      }
      // update anum
   	sprintf(sql, "delete from aa where anum = %d;", anum_initial); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	isc_print_status(args->status); 
      	goto cleanup;
      }
   
   
   	// try commit, but if it fails, do rollback
   	if (fbcommit(args, &tr1)) {
   		// commit failed...
   		isc_print_status(args->status);
   		if (fbrollback(args, &tr1)) {
   			isc_print_status(args->status);
   			printf ("\tMWRITE8 ROLLBACK #2 FAILED. THIS IS BAD.\n"); 		
   			goto cleanup;
   		}
   	}
   	
   cleanup:
   	if (tr1 !=0) {
   		if (fbrollback(args, &tr1)) {
   			isc_print_status(args->status);
   			printf ("\tMWRITE8 ROLLBACK #3 FAILED. THIS IS BAD.\n"); 		
   			goto cleanup;
   		}
   	}

} // outer for loop


   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

return 0; 
} // mwrite12




// testTransactionSerializabilityTwinsProblem()
int mwrite13 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *osqlda = NULL;
   XSQLVAR				*ovar;
   int anum_initial,bnum_initial,anum,bnum=0;
	int countTTT =0;
	ISC_STATUS_ARRAY status; 
	
   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
   for (row = 0; row < args->iterations; row++) {
   
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   
   	// initial data fill is:
   	// INSERT INTO TTT VALUES (1,'A');
      // INSERT INTO TTT VALUES (2,'A');
      // INSERT INTO TTT VALUES (3,'A');
      // INSERT INTO TTT VALUES (4,'A');
      // INSERT INTO TTT VALUES (5,'A');
   
   	sprintf(sql,"SELECT COUNT(*) FROM TTT WHERE ANUM = 3;");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		int count3 = *(int *) osqlda->sqlvar[0].sqldata;
     		if (count3 != 1) {
     			printf ("MWRITE13 ERROR: select count(*) from ttt where anum=3 <> 1. Exiting thread!\n");
     			goto rollback;
     			}
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("MWRITE13 ERROR: BAD FETCH. Exiting thread!\n");
   		isc_print_status (args->status);
   		goto rollback;
   	}
    
   	sprintf(sql, "DELETE FROM TTT WHERE ANUM = 3;"); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	// possible to have deadlock here
      	isc_print_status(args->status);
      	goto rollback;
      }

		// Sleep(1); // give other thread a chance to update
   
   	sprintf(sql,"INSERT INTO TTT VALUES (3,'A');");
      if (fbexeci(args, &db1, &tr1, sql)) {
      	// possible to have deadlock here
      	isc_print_status(args->status);
      	goto rollback;
      }
   
   	sprintf(sql,"SELECT COUNT(*) FROM TTT WHERE ANUM = 3;");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		int count3 = *(int *) osqlda->sqlvar[0].sqldata;
     		if (count3 != 1) {
     			printf ("MWRITE13 ERROR: select count(*) from ttt where anum=3 <> 1. Exiting thread!\n");
     			goto rollback;
     			}
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("MWRITE13 ERROR: BAD FETCH. Exiting thread!\n");
   		isc_print_status (args->status);
   		goto rollback;
   	}
    
//   	sprintf(sql, "DELETE FROM TTT WHERE ANUM = 3;"); 
//      if (fbexeci(args, &db1, &tr1, sql)) {
//      	// possible to have deadlock here
//      	isc_print_status(args->status);
//      	goto rollback;
//      }
//   
//   	sprintf(sql,"INSERT INTO TTT VALUES (3,'A');");
//      if (fbexeci(args, &db1, &tr1, sql)) {
//      	// possible to have deadlock here
//      	isc_print_status(args->status);
//      	goto rollback;
//      }
   
   	// try commit, but if it fails, do rollback
   	if (fbcommit(args, &tr1)) {
   		// commit failed...
   		isc_print_status(args->status);
   	}
   
   	
   rollback:
   	if (tr1 !=0) {
   		if (fbrollback(args, &tr1)) {
   			isc_print_status(args->status);
   			printf ("\tMWRITE13 ROLLBACK #3 FAILED. THIS IS BAD.\n"); 		
   			goto cleanup;
   		}
   	}

} // outer for loop

cleanup:
   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

return 0; 
} // mwrite13



int mwrite14 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *osqlda = NULL;
   XSQLVAR				*ovar;
   int anum_initial,bnum_initial,anum,bnum=0;
	int countTTT =0;
	ISC_STATUS_ARRAY status; 
	int i=0;
	int success=0;
	

   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
   // this one deadlocks often, so over-riding the # of loops
   for (row = 0; row < 5; row++) {
   	// initial data fill is:
   	// INSERT INTO TTT VALUES (1,'A');
      // INSERT INTO TTT VALUES (2,'A');
      // INSERT INTO TTT VALUES (3,'A');
      // INSERT INTO TTT VALUES (4,'A');
      // INSERT INTO TTT VALUES (5,'A');
   
   	sprintf(sql, "INSERT INTO TTT VALUES (3,'B');"); 
   	i=0;
   	success=0;
// used to be 3 attempts
   	  	if (fbbeginTran(args, &db1, &tr1))
      		fberror(args, "beginTran1");
      	if (fbexeci(args, &db1, &tr1, sql)) {
      		// this will fail because of the PK constraint
      		// isc_print_status(args->status); 
   			if (fbrollback(args, &tr1)) {
   				isc_print_status(args->status);
   				printf ("\tMWRITE13 ROLLBACK #3 FAILED. THIS IS BAD.\n");
               // Sleep(2);
           	}
         } 
   		else {
   			// the insert took this time. Commit!
   			success=1;
   			if (fbcommit(args, &tr1)) {
   				printf ("MWRITE14: inner commit failed.\n");
   				isc_print_status(args->status);
   				}
   			} // else
   		i++;
   
} // outer for loop

   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

return 0; 
} // mwrite14





// testTransactionsSerializableAssignSequentialKey()
int mwrite15 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *osqlda = NULL;
   XSQLVAR				*ovar;
   int anum_initial,bnum_initial,anum,bnum=0;
	int countTTT =0;
	ISC_STATUS_ARRAY status; 
	int max_anum=0;
	
   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
   for (row = 0; row < args->iterations; row++) {
   
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   
   	sprintf(sql,"SELECT Max(anum) from tt;");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		max_anum = *(int *) osqlda->sqlvar[0].sqldata;
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("MWRITE15 ERROR: BAD FETCH. Exiting thread!\n");
   		isc_print_status (args->status);
   		goto rollback;
   	}
    
   	sprintf(sql, "INSERT INTO TT VALUES (9,%d);", max_anum+1); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	// possible to have deadlock here
      	isc_print_status(args->status);
      	goto rollback;
      }

   	// try commit, but if it fails, do rollback
   	if (fbcommit(args, &tr1)) {
   		// commit failed...
   		isc_print_status(args->status);
   	}
   	
   rollback:
   	if (tr1 !=0) {
   		if (fbrollback(args, &tr1)) {
   			isc_print_status(args->status);
   			printf ("\tMWRITE15 ROLLBACK #3 FAILED. THIS IS BAD.\n"); 		
   			goto cleanup;
   		}
   	}

} // outer for loop

cleanup:
   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

return 0; 
} // mwrite15



int mwrite16 (Threadargs args) {
	int rc = 0;
   char              sql[256];
   int               row;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *osqlda = NULL;
   XSQLVAR				*ovar;
   int anum_initial,bnum_initial,anum,bnum=0;
	int countTTT =0;
	ISC_STATUS_ARRAY status; 
	int max_anum=0;
	
   // Connect and set up ...
   if (fbconnect(args, args->database, &db1))
       fberror(args, "connect");

	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   	
   for (row = 0; row < args->iterations; row++) {
   
     	if (fbbeginTran(args, &db1, &tr1))
      	fberror(args, "beginTran1");
   
   	sprintf(sql,"SELECT Max(anum) from tt;");
   	if (fbprepexec(args, &tr1, &stmt1, sql)) 
      	fberror (args->status, "mwrite7 - fbprepexec()"); 
    	if (fballocSqlda(args, &stmt1, &osqlda ))
   		fberror(args, "alloc sqlda");
    	while ((rc = fbfetch(status, &stmt1, osqlda)) == 0) {
   		max_anum = *(int *) osqlda->sqlvar[0].sqldata;
   		} // while rows to fetch
      if (fbfreeStmt(args, &stmt1, DSQL_close))
          fberror(args, "drop statetment");
   	if (rc != 100) {
   		printf ("MWRITE16 ERROR: BAD FETCH. Exiting thread!\n");
   		isc_print_status (args->status);
   		goto rollback;
   	}
    
   	sprintf(sql, "INSERT INTO TT VALUES (999,%d);", max_anum+1); 
      if (fbexeci(args, &db1, &tr1, sql)) {
      	// possible to have deadlock here
      	isc_print_status(args->status);
      	goto rollback;
      }

   	// try commit, but if it fails, do rollback
   	if (fbcommit(args, &tr1)) {
   		// commit failed...
   		isc_print_status(args->status);
   	}
   	
   rollback:
   	if (tr1 !=0) {
   		if (fbrollback(args, &tr1)) {
   			isc_print_status(args->status);
   			printf ("\tMWRITE16 ROLLBACK #3 FAILED. THIS IS BAD.\n"); 		
   			goto cleanup;
   		}
   	}

} // outer for loop

cleanup:
   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

	printf ("MWRITE%d thread %d done\n", args->testcase, args->threadID); 

return 0; 
} // mwrite16


void term15or16 (const char *database, Threadargs args, int expected_num_rows) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;
   char              sql[2048];
   int subjectID; 
   int i,row,od_row ; 
   
	// init
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");
	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");

	// get value of anum
	sprintf (sql, "SELECT count(*) from tt;");  
	if (fbprepexec(args, &tr1, &stmt1, sql))
		fberror(args, sql);
  	if (fballocSqlda(args, &stmt1, &sqlda ))
		fberror(args, "alloc sqlda");
 	while ((rc = fbfetch(args, &stmt1, sqlda)) == 0) {
  		int row_count = *(int *)sqlda->sqlvar[0].sqldata;
  		if (row_count != expected_num_rows) 
  			printf ("expected %d rows, but actually was %d rows!!!\n", expected_num_rows, row_count);
		} // while rows to fetch
	if (rc != 100)
		fberror(args, "fetch");                

	// all done
	if (fbcommit(args, &tr1)) {
       fberror(args, "commit");
	}

   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

}






int ThreadFn(Threadargs args)
{
	switch (args->testcase) {
		case 1: 
			mwrite1 (args); 
			break; 
		case 2: 
			mwrite2 (args); 
			break; 
		case 3: 
			mwrite3 (args); 
			break; 
		case 4: 
			mwrite4 (args); 
			break; 
		case 5: 
			mwrite5 (args); 
			break; 
		case 6: 
			mwrite6 (args); 
			break; 
		case 7: 
			mwrite7 (args); 
			break; 
		case 8: 
			mwrite8 (args); 
			break; 
		case 9: 
			mwrite9 (args); 
			break; 
		case 11: 
			mwrite11 (args); 
			break; 
		case 12: 
			mwrite12 (args); 
			break; 
		case 13: 
			mwrite13 (args); 
			break; 
		case 14: 
			mwrite14 (args); 
			break; 
		case 15: 
			mwrite15 (args); 
			break; 
		case 16: 
			mwrite16 (args); 
			break; 
		default:
			printf ("Invalid test case #%d\n", args->testcase); 
			exit(0); 
		} // switch
   args->done = 1;
   return 0;
} // ThreadFn


void init1or2 (const char *database, Threadargs args) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;
   char              sql[2048];
   int subjectID; 
   int i,row,od_row ; 
   
	printf("...connecting\n"); 
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");

	printf ("...creating table tt\n"); 
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   
   sprintf(sql, "create table tt (dollars numeric(4), anum numeric(4))"); 
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "commit after create index/table");


   // -------------------------------------
   // Populate tables
   printf ("...populating table tt. 50 rows.\n");
   if (fbbeginTran(args, &db1, &tr1))
       fberror(args, "beginTran");

   for (row = 1; row <= 50; row++) {
	   sprintf((char*)sql, "INSERT INTO tt VALUES (500, %d) ; ", row);
      if (fbexeci(args, &db1, &tr1, sql))
          fberror(args, "insert into tt");
      } // for row

   if (fbcommit(args, &tr1))
      fberror(args, "commit on insert into tables");
      
   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");
}

void init3or4 (const char *database, Threadargs args) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;
   char              sql[2048];
   int subjectID; 
   int i,row,od_row ; 
   
	printf("...connecting\n"); 
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   
	printf ("...creating table aa\n"); 
   sprintf(sql, "create table aa (ANUM integer)"); 
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "commit after create index/table");

	printf ("...creating table bb\n"); 
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   
   sprintf(sql, "create table bb (BNUM integer)"); 
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "commit after create index/table");

   // -------------------------------------
   // Populate tables
   printf ("...populating table aa. 1 row.\n");
	sprintf ((char*) sql, "insert into aa values (1) ; "); 
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into aa");
	
   printf ("...populating table bb. 1 row.\n");
	sprintf ((char*) sql, "insert into bb values (100) ; "); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into aa");

   if (fbcommit(args, &tr1))
      fberror(args, "commit on insert into tables");
      
   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");
}

void init5or6 (const char *database, Threadargs args) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   char              sql[2048];
   int i; 
   
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   
   sprintf(sql, "drop table aa;"); 
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "dropping table aa");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   sprintf(sql, "create table aa (ANUM integer)"); 
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "commit after create index/table");

   printf ("populating table aa for mwrite5/6.\n");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   for (i=0;i<25;i++) {
	sprintf ((char*) sql, "insert into aa values (1 + %d) ; ", i); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into aa");
   }
   if (fbcommit(args, &tr1))
      fberror(args, "commit after create index/table");
   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");
	
}

void init7or8 (const char *database, Threadargs args) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   char              sql[2048];
   int i; 
   
	
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");

   sprintf(sql, "drop table aa;"); 
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "drop table bb;"); 
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "drop table ttt;"); 
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "drop table tt;"); 
   fbexeci(args, &db1, &tr1, sql);
   
   if (fbcommit(args, &tr1))
      fberror(args, "dropping tables for init7or8");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   sprintf(sql, "create table aa (anum integer);"); 
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "CREATE TABLE BB (BNUM integer);");
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "CREATE TABLE TTT (ANUM integer NOT NULL UNIQUE, AUTHOR CHAR(1));");
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "CREATE TABLE TT (DOLLARS integer, ANUM integer);");
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "commit after create index/table");

   printf ("populating tables, init7or8()");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
	sprintf ((char*) sql, "INSERT INTO AA VALUES (1);"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "populating table aa");
   	
	sprintf ((char*) sql, "INSERT INTO BB VALUES (100);"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "populating table bb");
   	
   if (fbcommit(args, &tr1))
      fberror(args, "commit after create index/table");
   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");
}

void init9 (const char *database, Threadargs args) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   char              sql[2048];
   int i; 
   
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   
   sprintf(sql, "drop table test_blob;"); 
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "drop table onerow;"); 
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "drop generator test_blob_id_gen;"); 
   fbexeci(args, &db1, &tr1, sql);
   
   if (fbcommit(args, &tr1))
      fberror(args, "dropping table onerow");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   sprintf(sql, "create table test_blob (id integer not null primary key, obj_data blob, ts_field timestamp, t_field time);"); 
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "create generator test_blob_id_gen;");
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "create table onerow (i integer);");
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "commit after create index/table");

   printf ("populating tables for mwrite9.\n");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
	sprintf ((char*) sql, "insert into onerow values(1);"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into aa");
   if (fbcommit(args, &tr1))
      fberror(args, "commit after create index/table");
   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");
}

void init11or12 (const char *database, Threadargs args) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   char              sql[2048];
   int i; 
   
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   
   sprintf(sql, "drop table AA;"); 
   fbexeci(args, &db1, &tr1, sql);
   sprintf(sql, "drop table BB;"); 
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "init11or12: dropping tables");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   sprintf(sql, "CREATE TABLE AA (anum integer);"); 
   if (fbexeci(args, &db1, &tr1, sql))
  	    isc_print_status ( args->status) ; 
   sprintf(sql, "CREATE TABLE BB (bnum integer);");
   if (fbexeci(args, &db1, &tr1, sql))
  	    isc_print_status ( args->status) ; 
   if (fbcommit(args, &tr1))
      fberror(args, "init11or12: creating tables");

   printf ("populating tables for init11or12.\n");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
	sprintf ((char*) sql, "INSERT INTO AA VALUES (1)"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into aa");
	sprintf ((char*) sql, "INSERT INTO BB VALUES (6)"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into bb");
   if (fbcommit(args, &tr1))
      fberror(args, "init11or12: populating tables");
   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");
}

void init13or14 (const char *database, Threadargs args) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   char              sql[2048];
   int i; 
   
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   
   sprintf(sql, "drop table TTT;"); 
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "init13or14: dropping tables");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   sprintf(sql, "CREATE TABLE TTT (ANUM NUMERIC(4) NOT NULL unique, AUTHOR CHAR(1));"); 
   if (fbexeci(args, &db1, &tr1, sql))
  	    isc_print_status ( args->status) ; 
   if (fbcommit(args, &tr1))
      fberror(args, "init13or14: creating tables");

   printf ("populating tables for init13or14.\n");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");

	sprintf ((char*) sql, "INSERT INTO TTT VALUES (1,'A');"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into ttt");

	sprintf ((char*) sql, "INSERT INTO TTT VALUES (2,'A');"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into ttt");

	sprintf ((char*) sql, "INSERT INTO TTT VALUES (3,'A');"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into ttt");
   	
	sprintf ((char*) sql, "INSERT INTO TTT VALUES (4,'A');"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into ttt");
   	
	sprintf ((char*) sql, "INSERT INTO TTT VALUES (5,'A');"); 
   if (fbexeci(args, &db1, &tr1, sql))
   	fberror(args, "insert into ttt");

   if (fbcommit(args, &tr1))
      fberror(args, "init11or12: populating tables");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");
} // init13or14

void init15or16 (const char *database, Threadargs args) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   char              sql[2048];
   int i; 
   
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   
   sprintf(sql, "drop table TT;"); 
   fbexeci(args, &db1, &tr1, sql);
   if (fbcommit(args, &tr1))
      fberror(args, "init13or14: dropping tables");

   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");
   sprintf(sql, "CREATE TABLE TT (DOLLARS integer, ANUM integer);"); 
   if (fbexeci(args, &db1, &tr1, sql))
  	    isc_print_status ( args->status) ; 
   if (fbcommit(args, &tr1))
      fberror(args, "init15or16: creating tables");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");
} // init15or16

void queryTableAABB (const char *database, Threadargs args) {
   int rc;
   ISC_STATUS status;
   char dpb_buffer[256];
   char *dpb = dpb_buffer;
   isc_db_handle     db1 = {0};
   isc_tr_handle     tr1 = {0};
   isc_stmt_handle   stmt1 = {0};
   XSQLDA            *sqlda = NULL;
   char              sql[2048];
   int subjectID; 
   int i,row,od_row ; 
   
	// init
	if (fbconnect(args, database, &db1))
      fberror(args, "connect");
	if (fballocStmt(args, &db1, &stmt1))
   	fberror(args, "alloc statement");
   if (fbbeginTran(args, &db1, &tr1))
   	fberror(args, "beginTran");

	// get value of anum
	sprintf (sql, "SELECT anum FROM AA;");  
	if (fbprepexec(args, &tr1, &stmt1, sql))
		fberror(args, sql);
  	if (fballocSqlda(args, &stmt1, &sqlda ))
		fberror(args, "alloc sqlda");
 	while ((rc = fbfetch(args, &stmt1, sqlda)) == 0) {
  		int anum = *(int *)sqlda->sqlvar[0].sqldata;
		printf ("anum: %d\n", anum); 
		} // while rows to fetch
	if (rc != 100)
		fberror(args, "fetch");                
   if (fbfreeStmt(args, &stmt1, DSQL_close))
       fberror(args, "drop statetment");

	// get value of bnum
	sprintf (sql, "SELECT bnum FROM bb;");  
	if (fbprepexec(args, &tr1, &stmt1, sql))
		fberror(args, sql);
  	if (fballocSqlda(args, &stmt1, &sqlda ))
		fberror(args, "alloc sqlda");
 	while ((rc = fbfetch(args, &stmt1, sqlda)) == 0) {
  		int bnum = *(int *)sqlda->sqlvar[0].sqldata;
		printf ("bnum: %d\n", bnum); 
		} // while rows to fetch
	if (rc != 100)
		fberror(args, "fetch");                

	// all done
	fbfreeSqlda(args, &sqlda);
	if (fbcommit(args, &tr1)) {
       fberror(args, "commit");
	}

   if (fbfreeStmt(args, &stmt1, DSQL_drop))
       fberror(args, "drop statetment");

   if (fbdisconnect(args, &db1))
       fberror(args, "disconnect");

}


int main(int argc, char* argv[])
{
   ThreadArgs        args[500] = {0};
   fbLock            lock = {0};
   int               thread;
   long              start;
   long              end;
   int threads_per_test [NUM_TESTS+1] = {0};
   int argloop;
	int testcase;
	int num_threads;
	int run1or2 = 0; 
	int run3or4 = 0; 
	int run5or6 = 0; 
	int run7or8 = 0; 
	int run9 = 0;
	int run11or12 =0;
	int run13or14 =0; 
	int run15or16 =0; 


   if (argc < 3)
   {
      printf("Usage:\t%s mwrite[1-%d] #threads ... <db.fdb>\n\n", argv[0],NUM_TESTS);
      printf("mwrite1 - testTransactionsSerializableSelectUpdateWithArithmetic()\n"); 
      printf("mwrite2 - paired with mwrite1.\n"); 
      printf("mwrite3 - testTransactionSerializabilityDeadlockManagement()\n"); 
      printf("mwrite4 - paired with mwrite3.\n"); 
      printf("mwrite5 - serializablePhantomRead()\n"); 
      printf("mwrite6 - paired with mwrite5.\n");
      printf("mwrite7 - transactionRollbackSerialization().\n");
      printf("mwrite8 - paired with mwrite7.\n"); 
      printf("mwrite9 - generator/timestamp test w prepared statement.\n"); 
      printf("mwrite11 - testTransactionSerializabilityReadDeleteInsert().\n");
      printf("mwrite12 - paired with mwrite11.\n"); 
      printf("mwrite13 - testTransactionSerializabilityTwinsProblem().\n");
      printf("mwrite14 - paired with mwrite14.\n"); 
      printf("---\n"); 

		printf("\n"); 
      printf ("Example usage:\n");
      printf ("\t%s mwrite1 1 mwrite2 1 mw.fdb\n", argv[0]); 
      exit(1);
   }
   

// find the total number of threads that will be needed, and do minimal error checking
	num_threads=0;
	for (argloop=1;argloop<argc;argloop++) {
		int threads;
		int iterations;
		
		if ((strncmp("mwrite", argv[argloop], 6)==0)) {
			char testcase_s[256];
			strcpy (testcase_s, (char *) & argv[argloop][6]);
			testcase = atoi (testcase_s);
			if (testcase < 1 || testcase > NUM_TESTS) {
				printf ("Valid range for test cases is between 1 and %d.\n", NUM_TESTS); 
				exit(0); 
			}
			threads_per_test[testcase] = atoi(argv[argloop+1]);
			num_threads = num_threads + atoi(argv[argloop+1]);
			if ((testcase == 1) || (testcase==2)) 
				run1or2 = 1;
			if ((testcase == 3) || (testcase==4)) 
				run3or4 = 1;
			if ((testcase == 5) || (testcase==6)) 
				run5or6 = 1;
			if ((testcase == 7) || (testcase==8)) 
				run7or8 = 1;
			if ((testcase == 9) ) 
				run9 = 1;
			if ((testcase == 11) || (testcase==12))
				run11or12 = 1; 
			if ((testcase == 13) || (testcase==14))
				run13or14 = 1; 
			if ((testcase == 15) || (testcase==16))
				run15or16 = 1; 
		}		
	}
	
	// print pretty output report for later comparisons
	printf ("Testcase #Threads #Iterations\n");
	printf ("-------- -------- -----------\n"); 
	for (testcase=1;testcase <=NUM_TESTS; testcase++) {
		printf ("MWRITE%-2d      %3d        %4d\n", testcase, threads_per_test[testcase],NUM_LOOPS);  
	}
	printf ("-----------------------------\n"); 
	printf ("Total # of threads: %d\n", num_threads); 

	if (num_threads > 500) {
		printf ("ERROR: No more than 500 threads total may be specified.\n"); 
		exit(0); 
	}

	thread=0;
	for (testcase=1;testcase<=NUM_TESTS;testcase++) {
		int loop;
		for (loop=0;loop<threads_per_test[testcase];loop++) {
			args[thread].testcase=testcase;
			args[thread].iterations=NUM_LOOPS;
			thread++;
		}
	}
	printf ("%d threads initialized\n", thread); 
	
	
	if (run1or2) {
		printf ("initializing tables for mwrite1/2\n");
 		init1or2 (argv[argc-1], &args[0]);
	}
	
	if (run3or4) {
		printf ("Initial values for tables AA/BB\n");
		printf ("Expecting anum to be 1 and bnum to be 100\n"); 
 		init3or4 (argv[argc-1], &args[0]);
		queryTableAABB (argv[argc-1],&args[0]);
	}
	if (run5or6) {
		printf ("initializing tables for mwrite5/6\n");
 		init5or6 (argv[argc-1], &args[0]);
	}
	if (run7or8) {
		printf ("initializing tables for mwrite7/8\n");
 		init7or8 (argv[argc-1], &args[0]);
	}
	if (run9) {
		printf ("initializing tables for mwrite9\n");
 		init9 (argv[argc-1], &args[0]);
	}
	if (run11or12) {
		printf ("initializing tables for mwrite11/12\n");
 		init11or12 (argv[argc-1], &args[0]);
	}
	if (run13or14) {
		printf ("initializing tables for mwrite13/14\n");
 		init13or14 (argv[argc-1], &args[0]);
	}
	if (run15or16) {
		printf ("initializing tables for mwrite15/16\n");
 		init15or16 (argv[argc-1], &args[0]);
	}
	
   fbInitLock(&lock);

   start = GetTickCount();

   // Start requested number of threads
   for (thread = 0; thread < num_threads; thread++)
    {   
       // Pass a reference to the lock, since this API is not thread safe
       args[thread].plock = &lock;

       // Note: Using the same table increases concurrency issues
       // if (!SAME_TABLE)
           // args[thread].tableID = thread;

       args[thread].threadID = thread;

       // Tell thread the database to use
       // Note: We can use different databases to reduce concurrency issues
       args[thread].database = argv[argc-1];

       // Reset thread state
       args[thread].done = 0;

       // Start the thread
       fbCreateThread(ThreadFn, &args[thread]);

// Debug code:
//     // Wait for this thread to finish for full serialization
//     while (!args[thread].done)
//          Sleep(1);

   } // for thread

   printf("\nStarted %d threads ...\n", thread);

   // Wait for completion ...
   for (thread = 0; thread < num_threads; thread++)
   {
       while (!args[thread].done)
            Sleep(1);

   } // for thread

   end = GetTickCount();




	// printf ("Processed %d tests using %d threads in %f seconds.\n\n", num_threads*NUM_LOOPS, num_threads, (end-start)/1000.0 ); 
	printf ("The tests took %f seconds to complete.\n\n", (end-start)/1000.0); 

   fbDestroyLock(&lock);
   
//	if (run3or4) {
//		printf ("Final values for tables AA/BB\n");
//		printf ("Expecting anum to be 131 and bnum to be -30\n"); 
//		queryTableAABB (argv[argc-1],&args[0]);
//	}


	if (run15or16) {
		term15or16 (argv[argc-1],&args[0], (threads_per_test[15]*NUM_LOOPS)+(threads_per_test[16]*NUM_LOOPS));
	}

   return 0;
} // main
